/**
 * \file
 * \brief  !BETA! Library for creating deadline-based threads for the EDF scheduler
 *
 * \date   Aug 27th 2014
 * \author Valentin Hauner <vh@hnr.name>
 */
/* 
 * (c) 2014 Valentin Hauner <vh@hnr.name>
 *
 * This file is distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
#include <l4/edft/edft.h>

#include <stdio.h>
#include <stdlib.h>

Edf_thread     thread[THREAD_MAX_NUM];
l4_umword_t   *thread_stack[THREAD_MAX_NUM];

unsigned       count = 0;

int edft_no2index(unsigned no)
{
  unsigned i;
  for (i = 0; i < count; i++)
  {
    if (thread[i].no == no)
      return i;
  }
  return -1;
}

Edf_thread * edft_obj(unsigned no)
{
  return &thread[edft_no2index(no)];
}

void edft_release_cap(unsigned no)
{
  l4re_util_cap_free_um(thread[edft_no2index(no)].cap);
}

int edft_create_l4_thread(Edf_thread *t, unsigned run)
{
  l4_msgtag_t tag;

  l4_cap_idx_t task_cap = l4re_util_cap_alloc();
  // The second argument of l4_fpage denotes the size of the flex-page in log2
  l4_fpage_t task_fpage = l4_fpage(l4re_env()->first_free_utcb, 12, L4_CAP_FPAGE_RW);
  l4_factory_create_task(l4re_env()->factory, task_cap, task_fpage);
  l4_task_map(task_cap, L4RE_THIS_TASK_CAP,
               l4_obj_fpage(l4re_env()->rm, 0, L4_FPAGE_RO),
	       l4_map_obj_control(l4re_env()->rm, L4_MAP_ITEM_MAP));

  t->cap = l4re_util_cap_alloc();
  thread[count] = *t;

  thread_stack[count] = malloc(THREAD_STACK_SIZE * sizeof(l4_umword_t));
  // For x86-32, the arguments of functions are passed via the stack
  thread_stack[count][THREAD_STACK_SIZE - 1] = (l4_umword_t)t->arg;
  thread_stack[count][THREAD_STACK_SIZE - 2] = 0;

  if (l4_is_invalid_cap(t->cap))
    return -1;

  tag = l4_factory_create_thread(l4re_env()->factory, t->cap);
  if (l4_error(tag))
    return -1;

  l4_thread_control_start();
  l4_thread_control_pager(l4re_env()->rm);
  l4_thread_control_exc_handler(l4re_env()->rm);
  l4_thread_control_bind((l4_utcb_t *)l4re_env()->first_free_utcb,
                          task_cap);
//                          L4RE_THIS_TASK_CAP);
  tag = l4_thread_control_commit(t->cap);
  if (l4_error(tag))
    return -2;

  // Set the stack pointer to the second last element in the stack array for passing the arguments to the function
  tag = l4_thread_ex_regs(t->cap,
                          (l4_umword_t)t->func,
                          (l4_umword_t)&thread_stack[count][THREAD_STACK_SIZE - 2], 0);
  if (l4_error(tag))
    return -3;

  /*
   * The following mappings had no effect!
   */
  /*
  l4_task_map(task_cap, L4RE_THIS_TASK_CAP,
               l4_fpage((l4_umword_t)t, sizeof(t), L4_FPAGE_RW),
               l4_map_obj_control((l4_umword_t)&t, L4_MAP_ITEM_GRANT));
  l4_task_map(task_cap, L4RE_THIS_TASK_CAP,
               l4_obj_fpage(t->cap, 0, L4_FPAGE_RW),
               l4_map_obj_control(t->cap, L4_MAP_ITEM_GRANT));
  */

  // Immediate call of l4_scheduler_run_thread?
  if (run)
    if (edft_run_l4_thread(count) < 0)
      return -4;

  l4_debugger_set_object_name(t->cap, t->name);

  // Shift first_free_utcb for further threads
  l4re_env()->first_free_utcb = (l4_addr_t)l4re_env()->first_free_utcb + L4_UTCB_OFFSET;

  return count++;
}

int edft_run_l4_thread(unsigned index)
{
  l4_msgtag_t tag;

  // Pass the deadline of the thread to the L4 system
  l4_sched_param_t sp = l4_sched_param(1, 0); //l4_sched_param_by_type(Deadline, thread[index].dl, 0);
  // Let the L4 system tell the kernel to enqueue the thread in its (deadline-based) ready queue
  tag = l4_scheduler_run_thread(l4re_env()->scheduler, thread[index].cap, &sp);
  if (l4_error(tag))
    return -1;

  return 0;
}

int edft_run_l4_threads(void)
{
  unsigned i;
  int ret;

  for (i = 0; i < count; i++)
  {
    ret = edft_run_l4_thread(i);
    if (ret < 0)
      return ret;
  }

  return 0;
}

void edft_create_dependent_threads(Edf_thread *t, unsigned offset, unsigned run)
{
  // Calculate absolute deadline by adding current offset and worst case execution time
  t->dl = offset + t->wcet;
  edft_create_l4_thread(t, run);

  // Iterate through all successors of the thread by depth-first search
  Edf_suc *cur = t->suc;
  while (cur != NULL)
  {
    edft_create_dependent_threads(cur->data, offset + t->wcet, run);
    cur = cur->next;
  }
}

void edft_print_precedence_graph(Edf_thread *t, unsigned level)
{
  if (level == 0)
    printf(ANSI_BOLD "\n=== PRECEDENCE GRAPH ===\n" ANSI_RESET);

  printf("\n");

  unsigned i;
  for (i = 0; i < level; i++)
    printf("    ");
  for (i = 0; i < level; i++)
    printf("|");

  if (level > 0)
    printf("=> ");

  printf(ANSI_COLOR_BLUE    "#%u-%s"     ANSI_RESET "("
          ANSI_COLOR_YELLOW "wcet:%u" ANSI_RESET ";"
          ANSI_COLOR_RED    "dl:%u"   ANSI_RESET ") ",
          t->no,
          t->name,
          t->wcet,
          t->dl);
  Edf_suc *cur = t->suc;
  while (cur != NULL)
  {
    edft_print_precedence_graph(cur->data, level + 1);
    cur = cur->next;
  }

  if (level == 0)
    printf(ANSI_BOLD "\n\n========================\n\n" ANSI_RESET);
}
